
from __future__ import annotations
from typing import TYPE_CHECKING, Any, TypeVar, Optional, Mapping, Union, Callable, overload
if TYPE_CHECKING:
    from types import TracebackType
    from .auth.types import AuthorizationGrant
    from .http.types import RequestFiles
    from .http.payload import Payload
    from .core.http_client_ASYNC import HTTPClient
    from .types import JSON_ro

from .auth import Token
from .auth import grants
from .core import grants as core_grants
from .core.http_client_ASYNC import (
    RedditHTTPClient,
    build_reddit_http_client,
    build_reddit_http_client_from_access_token,
)
from .exceptions import raise_for_reddit_error, raise_for_non_json_response
from .http.util.json_loading import load_json_from_response
from .util.redditwarp_installed_client_credentials import get_redditwarp_client_id, get_device_id


class Client:
    _TSelf = TypeVar('_TSelf', bound='Client')

    @staticmethod
    def from_praw_config(section_name: str, *, filepath: Optional[str] = None) -> Client:
        from .util.praw_config_ASYNC import client_from_praw_config  # Avoid cyclic import
        return client_from_praw_config(section_name, filepath=filepath)

    @classmethod
    def from_http(cls: type[_TSelf], http: HTTPClient) -> _TSelf:
        self = cls.__new__(cls)
        self._init(http)
        return self

    @classmethod
    def from_access_token(cls: type[_TSelf], access_token: str) -> _TSelf:
        http = build_reddit_http_client_from_access_token(access_token)
        return cls.from_http(http)

    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, client_id: str, client_secret: str, /) -> None: ...
    @overload
    def __init__(self, client_id: str, client_secret: str, /, *, grant: AuthorizationGrant) -> None: ...
    @overload
    def __init__(self, client_id: str, client_secret: str, refresh_token: str, /) -> None: ...
    @overload
    def __init__(self, client_id: str, client_secret: str, username: str, password: str, /) -> None: ...
    def __init__(self, *creds: str, grant: Optional[AuthorizationGrant] = None) -> None:
        client_id = client_secret = ''
        n = len(creds)
        if n == 0:
            client_id = get_redditwarp_client_id()
            grant = core_grants.InstalledClientGrant(get_device_id())
        elif n == 2:
            client_id, client_secret = creds
            if grant is None:
                grant = grants.ClientCredentialsGrant()
        elif n == 3:
            client_id, client_secret, refresh_token = creds
            grant = grants.RefreshTokenGrant(refresh_token)
        elif n == 4:
            client_id, client_secret, username, password = creds
            grant = grants.ResourceOwnerPasswordCredentialsGrant(username, password)
        else:
            raise TypeError

        http = build_reddit_http_client(client_id, client_secret, grant)
        self._init(http)

    def _init(self, http: HTTPClient) -> None:
        self.http: HTTPClient = http
        self.last_value: Any = None

        # Delay heavy import till client instantiation
        # instead of library import.
        from .siteprocs.ASYNC import SiteProcedures
        self.p: SiteProcedures = SiteProcedures(self)

    async def __aenter__(self: _TSelf) -> _TSelf:
        return self

    async def __aexit__(self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        await self.close()
        return None

    async def close(self) -> None:
        await self.http.close()

    async def request(self,
        verb: str,
        url: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        data: Optional[Union[Mapping[str, str], bytes]] = None,
        json: JSON_ro = None,
        files: Optional[RequestFiles] = None,
        payload: Optional[Payload] = None,
        timeout: float = -2,
        follow_redirects: Optional[bool] = None,
        snub: Optional[Callable[[JSON_ro], None]] = raise_for_reddit_error,
    ) -> Any:
        json_data = None
        try:
            resp = await self.http.request(verb, url, params=params, headers=headers,
                    data=data, json=json, files=files, payload=payload,
                    timeout=timeout, follow_redirects=follow_redirects)

            if resp.data:
                try:
                    json_data = load_json_from_response(resp)
                except ValueError as cause:
                    try:
                        raise_for_non_json_response(resp)
                    except Exception as exc:
                        raise exc from cause
                    raise

                if snub is not None:
                    snub(json_data)

            resp.ensure_successful_status()
        finally:
            self.last_value = json_data
        return json_data

    def set_access_token(self, access_token: str) -> None:
        http = self.http
        if not isinstance(http, RedditHTTPClient):
            raise RuntimeError(f"self.http must be {RedditHTTPClient.__name__}")
        http.authorizer.set_token(Token(access_token))

    def set_user_agent(self, s: Optional[str]) -> None:
        ua = self.http.user_agent_base
        if s is not None:
            ua = f"{ua} Bot !-- {s}"
        self.http.set_user_agent(ua)

RedditClient = Client
Reddit = Client
RedditWarp = Client




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Mapping, Any
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.my_account_ASYNC import MyAccount
    from ...models.user_relationship import UserRelationship, FriendRelationship
    from ...models.trophy import Trophy
    from ...models.karma_breakdown_entry import KarmaBreakdownEntry
    from ...types import JSON_ro

from .pull_subreddits_ASYNC import PullSubreddits
from ...model_loaders.my_account_ASYNC import load_account
from ...model_loaders.user_relationship import load_user_relationship, load_friend_relationship
from ...model_loaders.karma_breakdown_entry import load_karma_breakdown_entry
from ...model_loaders.trophy import load_trophy
from ...util.base_conversion import to_base36
from ... import exceptions
from ...http import exceptions as http_exceptions

class AccountProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client
        self.pull_subreddits: PullSubreddits = PullSubreddits(client)
        ("""
            Pull subreddits the current user is a member of.
            """)

    async def fetch(self) -> MyAccount:
        """Fetch information about the currently logged in user.

        .. .RETURNS

        :rtype: :class:`~.models.my_account_ASYNC.MyAccount`

        .. .RAISES

        :raises redditwarp.exceptions.OperationException:
            There is no user context.
        """
        root = await self._client.request('GET', '/api/v1/me')
        if len(root) < 6:
            raise exceptions.RejectedResultException('no user context')
        return load_account(root, self._client)

    async def get_preferences(self) -> Mapping[str, Any]:
        """Retrieve the current user's preferences.

        .. .RETURNS

        :rtype: `Mapping`\\[`str`, `Any`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        return await self._client.request('GET', '/api/v1/me/prefs')

    async def set_preferences(self, prefs: Mapping[str, JSON_ro]) -> Mapping[str, Any]:
        """Update the current user's preferences.

        .. .PARAMETERS

        :param prefs:
            See the API docs for available preference settings:
            `<https://www.reddit.com/dev/api/#PATCH_api_v1_me_prefs>`_.
        :type prefs: `Mapping`\\[`str`, :obj:`~.types.JSON_ro`]

        .. .RETURNS

        :returns:
            The updated preferences, as you would get from :meth:`.get_preferences`.
        :rtype: `Mapping`\\[`str`, `Any`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        return await self._client.request('PATCH', '/api/v1/me/prefs', json=prefs)

    async def karma_breakdown(self) -> Sequence[KarmaBreakdownEntry]:
        """Get the current user's karma breakdown by subreddit.

        The entries are sorted in descending order by comment karma plus
        submission karma.

        .. .RETURNS

        :rtype: `Sequence`\\[:class:`~.models.karma_breakdown_entry.KarmaBreakdownEntry`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        root = await self._client.request('GET', '/api/v1/me/karma')
        entries = root['data']
        return [load_karma_breakdown_entry(d) for d in entries]

    async def trophies(self) -> Sequence[Trophy]:
        """Get a list of trophies for the current user.

        .. .RETURNS

        :rtype: `Sequence`\\[:class:`~.models.trophy.Trophy`]

        ..RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        root = await self._client.request('GET', '/api/v1/me/trophies')
        kind_data = root['data']['trophies']
        return [load_trophy(d['data']) for d in kind_data]

    async def list_friends(self) -> Sequence[UserRelationship]:
        """Get a list of friends.

        .. .RETURNS

        :rtype: `Sequence`\\[:class:`~.models.user_relationship.UserRelationship`]

        .. .RAISES

        :raises redditwarp.exceptions.OperationException:
            There is no user context.
        """
        try:
            root = await self._client.request('GET', '/api/v1/me/friends')

        except http_exceptions.StatusCodeException as e:
            if e.status_code == 302:
                raise exceptions.OperationException('no user context')
            raise
        if isinstance(root, str):
            raise exceptions.OperationException('no user context')

        entries = root['data']['children']
        return [load_user_relationship(d) for d in entries]

    async def get_friend(self, name: str) -> UserRelationship:
        """Get information about a specific 'friend'.

        This is an old reddit feature. See `<https://www.reddit.com/prefs/friends/>`_.

        .. .PARAMETERS

        :param `str` name:
            The name of a user.

        .. .RETURNS

        :rtype: :class:`~.models.user_relationship.UserRelationship`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NOT_FRIEND`:
                You are not friends with the specified user.
            + `USER_DOESNT_EXIST`:
                The specified user does not exist.
        """
        root = await self._client.request('GET', f'/api/v1/me/friends/{name}')
        return load_user_relationship(root)

    async def add_friend(self, name: str, note: Optional[str] = None) -> FriendRelationship:
        """Create or update a friend relationship.

        Also use this function to add or update a note.
        Note: making a note requires Reddit Premium.

        .. .PARAMETERS

        :param `str` name:
            The name of a user.
        :param `Optional[str]` note:
            Add or update a note.

            Using this parameter requires Reddit Premium, otherwise a
            `GOLD_REQUIRED` API error will occur.

        .. .RETURNS

        :returns:
            The updated user object on success.
        :rtype: :class:`~.models.user_relationship.FriendRelationship`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `USER_DOESNT_EXIST`:
                The specified user does not exist.
            + `GOLD_REQUIRED`:
                You tried to add a note but you don't have Reddit Premium.
            + `NO_TEXT`:
                An empty string was specified for `note`.
        """
        json_data = {} if note is None else {'note': note}
        root = await self._client.request('PUT', f'/api/v1/me/friends/{name}', json=json_data)
        return load_friend_relationship(root)

    async def remove_friend(self, name: str) -> None:
        """Remove a friend.

        .. .PARAMETERS

        :param `str` name:
            The name of a user.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `NOT_FRIEND`:
                The user specified is not a friend.
        """
        await self._client.request('DELETE', f'/api/v1/me/friends/{name}')

    async def list_blocked_users(self) -> Sequence[UserRelationship]:
        """Get a list of blocked users.

        .. .RETURNS

        :rtype: `Sequence`\\[:class:`~.models.user_relationship.UserRelationship`]

        .. .RAISES

        :raises redditwarp.exceptions.OperationException:
            There is no user context.
        """
        try:
            root = await self._client.request('GET', '/prefs/blocked')

        except http_exceptions.StatusCodeException as e:
            if e.status_code == 302:
                raise exceptions.OperationException('no user context')
            raise
        if isinstance(root, str):
            raise exceptions.OperationException('no user context')

        entries = root['data']['children']
        return [load_user_relationship(d) for d in entries]

    async def block_user_by_id(self, idn: int) -> None:
        """Block a user by ID.

        .. .PARAMETERS

        :param `int` idn:
            The ID of the user to block.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `400`:
               - The username or user ID given doesn't exist.
               - You tried to block yourself.
        """
        await self._client.request('POST', '/api/block_user', data={'account_id': to_base36(idn)})

    async def block_user_by_name(self, name: str) -> None:
        """Block a user by name.

        .. .PARAMETERS

        :param `str` name:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :(raises):
            (Same as in :meth:`.block_user_by_id`.)
        """
        await self._client.request('POST', '/api/block_user', data={'name': name})

    async def unblock_user_by_target_id(self, target_id: int, agent_id: int) -> None:
        """Unblock a user by ID.

        .. .PARAMETERS

        :param `int` target_id:
            The user ID in which to unblock.
        :param `int` agent_id:
            Your user account's ID.

            The endpoint needs this information for some dumb reason.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `400`:
                The target username or target user ID doesn't exist.
        """
        data = {
            'type': 'enemy',
            'container': 't2_' + to_base36(agent_id),
            'id': 't2_' + to_base36(target_id),
        }
        await self._client.request('POST', '/api/unfriend', data=data)

    async def unblock_user_by_target_name(self, target_name: str, agent_id: int) -> None:
        """Unblock a user by name.

        Behaves similarly to :meth:`.unblock_user_by_target_id`.

        .. .PARAMETERS

        :param `str` target_name:
            The username in which to unblock.
        :param `int` agent_id:
            Your user account's ID.

            The endpoint needs this information for some dumb reason.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :(raises):
            See :meth:`.unblock_user_by_target_id`.
        """
        data = {
            'type': 'enemy',
            'container': 't2_' + to_base36(agent_id),
            'name': target_name,
        }
        await self._client.request('POST', '/api/unfriend', data=data)

    async def list_trusted_users(self) -> Sequence[UserRelationship]:
        """Get a list of trusted users.

        Behaves similarly to :meth:`.blocked`.

        .. .RETURNS

        :rtype: `Sequence`\\[:class:`~.models.user_relationship.UserRelationship`]

        .. .RAISES

        :(raises):
            See :meth:`.blocked`.
        """
        try:
            root = await self._client.request('GET', '/prefs/trusted')

        except http_exceptions.StatusCodeException as e:
            if e.status_code == 302:
                raise exceptions.OperationException('no user context')
            raise
        if isinstance(root, str):
            raise exceptions.OperationException('no user context')

        entries = root['data']['children']
        return [load_user_relationship(d) for d in entries]

    async def add_trusted_user(self, name: str) -> None:
        """Add a user to your trusted users list.

        .. .PARAMETERS

        :param `str` name:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `CANT_WHITELIST_AN_ENEMY`:
                The specified user is on your blocked list.
            + `USER_DOESNT_EXIST`:
                The specified user does not exist.
        """
        await self._client.request('POST', '/api/add_whitelisted', params={'name': name})

    async def remove_trusted_user(self, name: str) -> None:
        """Remove a user from your trusted users list.

        .. .PARAMETERS

        :param `str` name:
            The target user name.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        await self._client.request('POST', '/api/remove_whitelisted', params={'name': name})

    async def get_blocked_and_trusted_users(self) -> tuple[Sequence[UserRelationship], Sequence[UserRelationship]]:
        """Return the blocked and trusted user lists in one call.

        Behaves similarly to :meth:`.blocked`.

        .. .RETURNS

        :returns:
            A tuple of two lists:
            the first list contains blocked users,
            the second list contains trusted's users.
        :rtype: `tuple`\\[`Sequence`\\[:class:`~.models.user_relationship.UserRelationship`], `Sequence`\\[:class:`~.models.user_relationship.UserRelationship`]]

        .. .RAISES

        :(raises):
            See :meth:`.blocked`.
        """
        try:
            root = await self._client.request('GET', '/prefs/messaging')

        except http_exceptions.StatusCodeException as e:
            if e.status_code == 302:
                raise exceptions.OperationException('no user context')
            raise
        if isinstance(root, str):
            raise exceptions.OperationException('no user context')

        blocked_entries = root[0]['data']['children']
        trusted_entries = root[1]['data']['children']
        return (
            [load_user_relationship(d) for d in blocked_entries],
            [load_user_relationship(d) for d in trusted_entries],
        )




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any, Iterable
if TYPE_CHECKING:
    from ..client_SYNC import Client
    from ..iterators.call_chunk_calling_iterator import CallChunkCallingIterator

from .custom_feed import CustomFeed as BaseCustomFeed

class CustomFeed(BaseCustomFeed):
    def __init__(self, d: Mapping[str, Any], client: Client) -> None:
        super().__init__(d)
        self.client: Client = client
        ("")

    def delete(self) -> None:
        self.client.p.custom_feed.delete(self.owner, self.name)

    def contains(self, sr_name: str) -> bool:
        return self.client.p.custom_feed.contains(self.owner, self.name, sr_name)

    def add_item(self, sr_name: str) -> None:
        self.client.p.custom_feed.add_item(self.owner, self.name, sr_name)

    def bulk_add_items(self, sr_names: Iterable[str]) -> CallChunkCallingIterator[None]:
        return self.client.p.custom_feed.bulk_add_items(self.owner, self.name, sr_names)

    def remove_item(self, sr_name: str) -> None:
        self.client.p.custom_feed.remove_item(self.owner, self.name, sr_name)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Mapping
if TYPE_CHECKING:
    from collections.abc import Iterator

# https://pypi.org/project/websocket-client/
import websocket  # type: ignore[import]

from ..reg_SYNC import register
from ...websocket_SYNC import PulsePartiallyImplementedWebSocketConnection, DEFAULT_WAITTIME
from ... import exceptions
from ... import events
from ...events import Frame
from ...const import Opcode, Side, ConnectionState
from ...utils import parse_close


def _get_necessary_timeout(timeout: float = -2) -> Optional[float]:
    t: Optional[float] = timeout
    if timeout == -2:
        t = DEFAULT_WAITTIME
    elif timeout == -1:
        t = None
    elif timeout < 0:
        raise ValueError(f'invalid timeout value: {timeout}')
    return t


class WebSocketClient(PulsePartiallyImplementedWebSocketConnection):
    SIDE: int = Side.CLIENT

    def __init__(self, ws: websocket.WebSocket) -> None:
        super().__init__()
        self.ws: websocket.WebSocket = ws
        ("")

    def _send_frame_impl(self, m: Frame) -> None:
        frm = websocket.ABNF.create_frame(opcode=m.opcode, data=m.data, fin=int(m.fin))
        try:
            self.ws.send_frame(frm)
        except websocket.WebSocketConnectionClosedException as cause:
            if self.state != ConnectionState.CLOSED:
                raise Exception('assertion')
            raise exceptions.ConnectionClosedException from cause
        except Exception as cause:
            raise exceptions.TransportError from cause

    def _load_next_frame(self, *, timeout: float = -2) -> Optional[Frame]:
        self.ws.timeout = _get_necessary_timeout(timeout)
        try:
            _, frm = self.ws.recv_data_frame(True)
        except websocket.WebSocketTimeoutException as cause:
            raise exceptions.TimeoutException from cause
        except websocket.WebSocketConnectionClosedException as cause:
            if self.state != ConnectionState.CLOSED:
                raise Exception('assertion')
            raise exceptions.ConnectionClosedException from cause
        except websocket.WebSocketProtocolException as cause:
            if "Invalid close opcode" in str(cause):
                self.close_code = cause.args[1]
                return None
            raise
        except Exception as cause:
            raise exceptions.TransportError from cause

        return Frame(
            opcode=Opcode(frm.opcode),
            fin=bool(frm.fin),
            data=(frm.data if isinstance(frm.data, bytes) else frm.data.encode()),
        )

    def _process_ping_frame(self, m: Frame) -> Iterator[object]:
        yield m

    def _process_close_frame(self, m: Frame) -> Iterator[object]:
        yield m
        self.close_code: int
        self.close_reason: str
        self.close_code, self.close_reason = parse_close(m.data)
        self.state = ConnectionState.CLOSED
        yield events.ConnectionClosed()

    def _send_close_frame_impl(self, code: Optional[int] = 1000, reason: str = '') -> None:
        try:
            self.ws.send_close(code, reason.encode())
        except Exception as cause:
            raise exceptions.TransportError from cause

    def close(self, code: Optional[int] = 1000, reason: str = '', *, waitfor: float = -2) -> None:
        super().close(code, reason, waitfor=waitfor)
        try:
            self.ws.shutdown()
        except Exception as cause:
            raise exceptions.TransportError from cause


def connect(
    url: str,
    *,
    subprotocols: Sequence[str] = (),
    headers: Optional[Mapping[str, str]] = None,
    timeout: float = -2,
) -> WebSocketClient:
    t = _get_necessary_timeout(timeout)
    try:
        ws = websocket.create_connection(
            url,
            fire_cont_frame=True,
            timeout=t,
            suppress_origin=True,
            header=headers,
            subprotocols=subprotocols,
        )
    except websocket.WebSocketTimeoutException as cause:
        raise exceptions.TimeoutException from cause
    except Exception as cause:
        raise exceptions.TransportError from cause
    wsc = WebSocketClient(ws)
    wsc.subprotocol = ws.subprotocol or ''
    return wsc


name: str = 'websocket-client'
version: str = websocket.__version__
register(
    adaptor_module_name=__name__,
    name=name,
    version=version,
    connect=connect,
)



"""A Pushshift API client.

Use this package as a scalpel. It's pagination based and doesn't do concurrency.

If you need to scrape lots of data, you'll be better using PMAW_.

.. _PMAW: https://github.com/mattpodolak/pmaw
"""

from . import core  # noqa: F401
from . import models  # noqa: F401
from . import paginators  # noqa: F401




from __future__ import annotations
from typing import Optional, ClassVar

from dataclasses import dataclass

from ..auth.grants import AuthorizationGrant

@dataclass(frozen=True)
class InstalledClientGrant(AuthorizationGrant):
    GRANT_TYPE: ClassVar[str] = "https://oauth.reddit.com/grants/installed_client"
    device_id: str
    scope: Optional[str] = None




from typing import Mapping, Any

from dataclasses import dataclass

from .datamemento import DatamementoDataclassesMixin

@dataclass(repr=False, eq=False)
class SubredditStyleAssetUploadLease(DatamementoDataclassesMixin):
    d: Mapping[str, Any]
    endpoint: str
    fields: Mapping[str, str]
    s3_object_key: str
    location: str



pass
'''
https://www.reddit.com/r/hacking/comments/zo43jm/good_reddit/

https://github.com/reddit-archive/reddit/blob/master/r2/r2/lib/signing.py
'''

import hmac
import hashlib


MOBILE_ANDROID_HMAC_KEY: str = "8c7abaa5f905f70400c81bf3a1a101e75f7210104b1991f0cd5240aa80c4d99d"
MOBILE_IOS_HMAC_KEY: str = ""  # Unknown

VENDOR_HEADER_NAME: str = 'client-vendor-id'
SIGNATURE_UA_HEADER_NAME: str = 'X-hmac-signed-result'
SIGNATURE_BODY_HEADER_NAME: str = 'X-hmac-signed-body'

SIGNATURE_UA_MESSAGE_TEMPLATE: str = "Epoch:%d|User-Agent:%s|Client-Vendor-ID:%s"
SIGNATURE_BODY_MESSAGE_TEMPLATE: str = "Epoch:%d|Body:%s"

SIGNATURE_TEMPLATE: str = "{global_version}:{platform}:{version}:{epoch}:{hmac_hash}"

VALID_EPOCH_MAX_AGE: float = 5 * 60


def generate_hmac_hash(secret: str, message: str) -> str:
    return hmac.new(secret.encode(), message.encode(), hashlib.sha256).hexdigest()

def generate_android_mobile_hmac_hash(message: str) -> str:
    return generate_hmac_hash(MOBILE_ANDROID_HMAC_KEY, message)

def get_android_mobile_request_signature(epoch: int, hmac_hash: str) -> str:
    return SIGNATURE_TEMPLATE.format_map({
        'global_version': 1,
        'platform': 'android',
        'version': 2,
        'epoch': epoch,
        'hmac_hash': hmac_hash,
    })




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping
if TYPE_CHECKING:
    from ..client_SYNC import Client

from ..models.custom_feed_SYNC import CustomFeed

def load_custom_feed(d: Mapping[str, Any], client: Client) -> CustomFeed:
    return CustomFeed(d, client)




from __future__ import annotations

from .listing.mixins.subreddit_detail_SYNC import SubredditDetail
from .listing.subreddit_listing_paginator import SubredditListingPaginator
from ...models.subreddit_SYNC import Subreddit

class SubscribedListingPaginator(
    SubredditDetail[Subreddit],
    SubredditListingPaginator,
): pass

class ContributingListingPaginator(
    SubredditDetail[Subreddit],
    SubredditListingPaginator,
): pass

class ModeratingListingPaginator(
    SubredditDetail[Subreddit],
    SubredditListingPaginator,
): pass




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ...http.http_client_ASYNC import HTTPClient as BaseHTTPClient

from ...core.http_client_ASYNC import HTTPClient
from ...http.transport.reg_ASYNC import new_connector
from ...http.util.case_insensitive_dict import CaseInsensitiveDict
from ...core.user_agent_ASYNC import get_user_agent
from .const import RESOURCE_BASE_URL
from .rate_limited_ASYNC import RateLimited

def build_http_client() -> BaseHTTPClient:
    connector = new_connector()
    ua = get_user_agent(module_member=connector)
    headers = CaseInsensitiveDict({'User-Agent': ua})
    http = HTTPClient(RateLimited(connector), headers=headers)
    http.base_url = RESOURCE_BASE_URL
    return http




from __future__ import annotations
from typing import TypeVar, AsyncIterator, Iterator, Generic, Optional, Sequence

from .async_paginator import AsyncPaginator

E = TypeVar('E')

class PaginatorChainingAsyncIterator(AsyncIterator[E]):
    def __init__(self, paginator: AsyncPaginator[E], amount: Optional[int] = None) -> None:
        self._paginator: AsyncPaginator[E] = paginator
        self._paginator_iterator: AsyncIterator[Sequence[E]] = paginator.__aiter__()
        self.remaining: Optional[int] = amount
        ("")
        self.current_iterator: Iterator[E] = iter(())
        ("")

    def __aiter__(self) -> AsyncIterator[E]:
        return self

    async def __anext__(self) -> E:
        limit = self._paginator.limit
        remaining = self.remaining
        if remaining is None or remaining > 0:
            while True:
                for elem in self.current_iterator:
                    if self.remaining is not None:
                        self.remaining -= 1
                    return elem

                if (limit is not None and remaining is not None) and limit > remaining:
                    self._paginator.limit = remaining

                it = await self._paginator_iterator.__anext__()
                self.current_iterator = iter(it)

        raise StopAsyncIteration



__bound = 'AsyncPaginator[E]'
TAsyncPaginator = TypeVar('TAsyncPaginator', bound=AsyncPaginator)  # type: ignore[type-arg]

class ImpartedPaginatorChainingAsyncIterator(PaginatorChainingAsyncIterator[E], Generic[TAsyncPaginator, E]):
    def __init__(self, paginator: TAsyncPaginator, amount: Optional[int] = None) -> None:
        super().__init__(paginator, amount)
        self.__paginator: TAsyncPaginator = paginator

    def get_paginator(self) -> TAsyncPaginator:
        return self.__paginator




from __future__ import annotations
from typing import Sequence

from ..listing.listing_paginator import ListingPaginator
from ....model_loaders.submission_SYNC import load_submission
from ....models.submission_SYNC import Submission

class SubmissionListingPaginator(ListingPaginator[Submission]):
    def fetch(self) -> Sequence[Submission]:
        data = self._fetch_data()
        return [load_submission(d['data'], self.client) for d in data['children']]




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Optional, Mapping, MutableMapping

from base64 import b64encode
from urllib.parse import urlencode

def ___authorization_url(  # UNUSED: scrapped idea
    url: str,
    response_type: str,
    client_id: str,
    redirect_uri: Optional[str],
    scope: Optional[str] = None,
    state: Optional[str] = None,
    extra_params: Optional[Mapping[str, str]] = None,
) -> str:
    params = {
        'response_type': response_type,
        'client_id': client_id,
        'redirect_uri': redirect_uri,
        'scope': scope,
        'state': state,
        **(extra_params or {}),
    }
    params = {k: v for k, v in params.items() if v}
    return '%s?%s' % (url, urlencode(params))

def basic_auth(userid: str, password: str, *,
        scheme: str = 'Basic') -> str:
    """Make a basic auth string suitable for use in an `Authorization` header."""
    b64_user_pass = b64encode(f"{userid}:{password}".encode()).decode()
    return f"{scheme} {b64_user_pass}"

def apply_basic_auth(
    headers: MutableMapping[str, str],
    userid: str,
    password: str,
    *,
    scheme: str = 'Basic',
    header_name: str = 'Authorization',
) -> None:
    """Write a basic auth string into the `Authorization` field of a headers mapping."""
    headers[header_name] = basic_auth(userid, password, scheme=scheme)




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .send_params import SendParams
    from .exchange import Exchange

from .delegating_handler_ASYNC import DelegatingHandler


class Invoker(DelegatingHandler):
    async def send(self, p: SendParams) -> Exchange:
        return await self._send(p)

    async def close(self) -> None:
        await self._close()




from __future__ import annotations
from typing import Mapping, Any, Optional, Sequence

from datetime import datetime, timezone

from ..core.const import AUTHORIZATION_BASE_URL
from .datamemento import DatamementoBase
from .report import ModReport, UserReport
from ..model_loaders.report import load_mod_report, load_user_report

class Comment(DatamementoBase):
    @property
    def submission(self) -> Comment.Submission:
        """Information related to the comment's submission."""
        return self.__submission

    @property
    def subreddit(self) -> Comment.Subreddit:
        """Information related to the comment's subreddit."""
        return self.__subreddit

    class Me:
        def __init__(self, d: Mapping[str, Any]) -> None:
            self.saved: bool = d['saved']
            ("""
                Whether the current user has saved the comment.

                Value false if there is no user context.
                """)
            self.reply_notifications: bool = d['send_replies']
            ("""
                Whether an inbox message will be sent to you when the submission receives a new top-level comment.

                Value true if you are not the author of the submission.

                Value true if there is no user context.
                """)
            self.voted: int = {False: -1, None: 0, True: 1}[d['likes']]
            ("""
                Value `1` if upvoted, `0` if not voted on, `-1` if downvoted.
                """)

    class Author:
        class AuthorFlair:
            def __init__(self, d: Mapping[str, Any]) -> None:
                flair_text: Optional[str] = d['author_flair_text']
                flair_css_class: Optional[str] = d['author_flair_css_class']

                self.template_uuid: Optional[str] = x if (x := d['author_flair_template_id']) else None
                ("""
                    The author's flair template UUID.

                    Value is null if no flair template is being used.
                    """)
                self.text_mode: str = d['author_flair_type']
                ("""
                    Either `text` or `richtext`.
                    """)
                self.text: str = flair_text or ''
                ("""
                    Flair text.

                    Check if the value is an empty string to tell if a flair is being used.
                    """)
                self.css_class: str = flair_css_class or ''
                ("""
                    The author's flair CSS class.

                    When a flair template is being used, the value of this field will be
                    that of the CSS class designated by the template. If the flair template
                    does not specify a CSS class then the value will be an empty string.
                    """)
                self.bg_color: str = '' if (x := d['author_flair_background_color']) == 'transparent' else x
                ("""
                    A background color hex string. E.g., `#46d160`.

                    If a flair template is not being used then the value is an empty string.

                    If a flair template is being used and the background color is unset then
                    the value is an empty string.
                    """)
                self.fg_color_scheme: str = d['author_flair_text_color'] or ''
                ("""
                    Either `dark` or `light`, or empty string.

                    Value is empty string if a flair template is not being used.
                    """)

                self.has_had_flair_assigned_before_in_subreddit: bool = flair_text is not None
                ("""
                    Because of quirks in the API, we can tell if the user has ever had a flair assigned before in the subreddit.
                    """)
                self.has_had_flair_css_class_assigned_before_in_subreddit_when_no_flair_template_assigned: bool = flair_css_class is not None
                ("""
                    Because of quirks in the API, we can tell if the user has ever had a flair CSS class
                    assigned before in the subreddit. This only works if a flair template is not being used.
                    """)

        def __init__(self, d: Mapping[str, Any]) -> None:
            self.name: str = d['author']
            ("")
            self.id36: str = d['author_fullname'].split('_', 1)[-1]
            ("")
            self.id: int = int(self.id36, 36)
            ("")
            self.has_premium: bool = d['author_premium']
            ("")
            self.flair: Comment.Author.AuthorFlair = self.AuthorFlair(d)
            ("""
                Attributes related to the author's flair.
                """)

    class Submission:
        def __init__(self, d: Mapping[str, Any]) -> None:
            self.id36: str = d['link_id'].split('_', 1)[-1]
            ("")
            self.id: int = int(self.id36, 36)
            ("")
            self.archived: bool = d['archived']
            ("""
                Whether the post is archived.

                Archived posts cannot be commented on,
                but the author can still edit the OP.
                """)

    class Subreddit:
        def __init__(self, d: Mapping[str, Any]) -> None:
            self.id36: str = d['subreddit_id'].split('_', 1)[-1]
            ("")
            self.id: int = int(self.id36, 36)
            ("")
            self.name: str = d['subreddit']
            ("")
            self.openness: str = d['subreddit_type']
            ("""
                One of: `public`, `private`, `restricted`, `archived`,
                `employees_only`, `gold_only`, `gold_restricted`, or `user`.
                """)

    class Moderation:
        class Approved:
            def __init__(self, d: Mapping[str, Any]) -> None:
                self.by: str = d['approved_by']
                ("""
                    Name of the moderator who approved this comment.
                    """)
                self.ut: int = d['approved_at_utc']
                ("")
                self.at: datetime = datetime.fromtimestamp(self.ut, timezone.utc)
                ("")

        class Removed:
            def __init__(self, d: Mapping[str, Any]) -> None:
                self.by: str = d['banned_by']
                ("""
                    Name of the moderator who removed this comment.
                    """)
                self.ut: int = d['banned_at_utc']
                ("")
                self.at: datetime = datetime.fromtimestamp(self.ut, timezone.utc)
                ("")

        class Reports:
            def __init__(self, d: Mapping[str, Any]) -> None:
                self.ignoring: bool = d['ignore_reports']
                ("")
                self.num_reports: int = d['num_reports']
                ("")
                self.mod_reports: Sequence[ModReport] = [load_mod_report(m) for m in d['mod_reports']]
                ("")
                self.user_reports: Sequence[UserReport] = [load_user_report(m) for m in d['user_reports']]
                ("")

        class RemovalReason:
            def __init__(self, d: Mapping[str, Any]) -> None:
                self.by: str = d['mod_reason_by'] or ''
                ("")
                self.title: str = d['mod_reason_title'] or ''
                ("")
                self.note: str = d['mod_note'] or ''
                ("")

        def __init__(self, d: Mapping[str, Any]) -> None:
            self.spam: bool = d['spam']
            ("")

            self.approved: Optional[Comment.Moderation.Approved] = None
            ("")
            if d['approved_by']:
                self.approved = self.Approved(d)

            self.removed: Optional[Comment.Moderation.Removed] = None
            ("")
            if d['banned_by']:
                self.removed = self.Removed(d)

            self.reports: Comment.Moderation.Reports = self.Reports(d)
            ("")

            self.removal_reason: Optional[Comment.Moderation.RemovalReason] = None
            ("")
            if d['mod_reason_by']:
                self.removal_reason = self.RemovalReason(d)

    class Edited:
        def __init__(self, outer: Comment) -> None:
            self.ut: int = outer.edited_ut
            ("")
            self.at: datetime = outer.edited_at
            ("")

    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.id36: str = d['id']
        ("""
            ID of the comment as a base 36 number.
            """)
        self.id: int = int(self.id36, 36)
        ("")
        self.created_ut: int = int(d['created_utc'])
        ("""
            Unix timestamp of when the comment was made.
            """)
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)
        ("""
            When the comment was made.
            """)

        self.body: str = d['body']
        ("""
            Body text of the comment. In markdown format.
            """)
        self.body_html: str = d.get('body_html', '')
        ("""
            HTML of the comment.
            """)
        self.score: int = d['score']
        ("""
            Number of upvotes (minus downvotes).
            """)
        self.score_hidden: bool = d['score_hidden']
        ("""
            Whether the score should be hidden.
            """)

        self.permalink_path: str = d['permalink']
        ("""
            URL of the comment, without the domain.
            """)
        self.permalink: str = AUTHORIZATION_BASE_URL + self.permalink_path
        ("""
            URL of the comment.
            """)

        edited: Any = d['edited']
        self.is_edited: bool = bool(edited)
        ("""
            True if the comment was edited.
            """)
        self.edited_ut: int = int(edited) if edited else 0
        ("""
            Unix timestamp of when the comment was edited.

            Value is `0` if :attr:`is_edited` is false.
            """)
        self.edited_at: datetime = datetime.min
        ("""
            When the comment was edited.

            Value is `datetime.min` if :attr:`is_edited` is false.
            """)
        if self.is_edited:
            self.edited_at = datetime.fromtimestamp(self.edited_ut, timezone.utc)

        self.edited: Optional[Comment.Edited] = None
        ("""
            Value non-`None` if the comment was edited.
            """)
        if edited:
            self.edited = self.Edited(self)

        self.is_submitter: bool = d['is_submitter']
        ("""
            True if the author of this comment is the submission author ("OP").
            """)
        self.stickied: bool = d['stickied']
        ("")
        self.locked: bool = d['locked']
        ("")
        self.collapsed: bool = d['collapsed']
        ("""
            Whether the comment is collapsed by default, i.e., when it has been downvoted significantly.
            """)
        self.distinguished: str = d['distinguished'] or ''
        ("""
            Empty string if not distinguished, otherwise `'moderator'` or `'admin'`.
            """)

        parent_id: str = d['parent_id']
        self.is_top_level: bool = parent_id.startswith('t3_')
        ("""
            Whether the comment is a direct child of the submission.
            """)
        self.has_parent_comment: bool = not self.is_top_level
        ("""
            Same as `not self.is_top_level`.
            """)
        self.parent_comment_id36: str = ''
        ("""
            Parent comment ID36.

            Empty string if not applicable.
            """)
        self.parent_comment_id: int = 0
        ("""
            Parent comment ID.

            Value is `0` if not applicable.
            """)
        if self.has_parent_comment:
            self.parent_comment_id36 = parent_id.partition('_')[2]
            self.parent_comment_id = int(self.parent_comment_id36, 36)

        self.me: Comment.Me = self.Me(d)
        ("""
            Attributes relating to the current user.

            If there is no user context, these values contain nonsense.
            """)

        self.__submission: Comment.Submission = self.Submission(d)
        self.__subreddit: Comment.Subreddit = self.Subreddit(d)

        s: str = d['author']
        self.author_display_name: str = s
        ("""
            The author's username.

            Possibly `[removed]` if the comment was removed or
            `[deleted]` if the comment was deleted by the author.
            """)
        self.author: Optional[Comment.Author] = None
        ("""
            Information about the author.

            Value is `None` if the comment was removed or deleted.
            """)
        if not s.startswith('['):
            self.author = self.Author(d)

        self.mod: Optional[Comment.Moderation] = None
        ("""
            Attributes relating to moderation.

            Value is `None` if the current user is not a moderator of the subreddit.
            """)
        if 'spam' in d:
            self.mod = self.Moderation(d)


class LooseComment(Comment):
    # For:
    # * `GET /comments`
    # * `GET /r/{subreddit}/comments`
    # * `GET /user/{username}/overview` (and variants)

    @property
    def submission(self) -> LooseComment.Submission:
        return self.__submission

    @property
    def subreddit(self) -> LooseComment.Subreddit:
        return self.__subreddit

    class Submission(Comment.Submission):
        def __init__(self, d: Mapping[str, Any]) -> None:
            super().__init__(d)
            self.title: str = d['link_title']
            ("")
            self.author_display_name: str = d['link_author']
            ("")
            self.permalink_path: str = d['link_permalink']
            ("")
            self.permalink: str = AUTHORIZATION_BASE_URL + self.permalink_path
            ("")
            self.nsfw: bool = d['over_18']
            ("")

    class Subreddit(Comment.Subreddit):
        def __init__(self, d: Mapping[str, Any]) -> None:
            super().__init__(d)
            self.quarantined: bool = d['quarantine']
            ("")

    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.__submission: LooseComment.Submission = self.Submission(d)
        self.__subreddit: LooseComment.Subreddit = self.Subreddit(d)




from __future__ import annotations
from typing import Mapping, Any, Iterator

from dataclasses import dataclass

from .datamemento import DatamementoDataclassesMixin


@dataclass(repr=False, eq=False)
class FlairEmojiUploadLease(DatamementoDataclassesMixin):
    d: Mapping[str, Any]
    endpoint: str
    fields: Mapping[str, str]
    s3_object_key: str
    location: str

@dataclass(repr=False, eq=False)
class FlairEmoji:
    d: Mapping[str, Any]
    name: str
    image_url: str
    creator_id: int
    creator_id36: str
    post_enabled: bool
    user_enabled: bool
    mod_only: bool


class SubredditFlairEmojis(Mapping[str, FlairEmoji]):
    def __init__(self,
        *,
        subreddit_emojis: Mapping[str, FlairEmoji],
        reddit_emojis: Mapping[str, FlairEmoji],
        all_emojis: Mapping[str, FlairEmoji],
        subreddit_id36: str,
    ) -> None:
        self.subreddit_emojis: Mapping[str, FlairEmoji] = subreddit_emojis
        ("")
        self.reddit_emojis: Mapping[str, FlairEmoji] = reddit_emojis
        ("")
        self.all_emojis: Mapping[str, FlairEmoji] = all_emojis
        ("")
        self.subreddit_id36: str = subreddit_id36
        ("")
        self.subreddit_id: int = int(subreddit_id36, 36)
        ("")

    def __contains__(self, item: object) -> bool:
        return item in self.all_emojis

    def __iter__(self) -> Iterator[str]:
        return iter(self.all_emojis)

    def __len__(self) -> int:
        return len(self.all_emojis)

    def __getitem__(self, key: str) -> FlairEmoji:
        return self.all_emojis[key]




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any, Sequence, Optional
if TYPE_CHECKING:
    from ..client_ASYNC import Client

from dataclasses import dataclass

from .modmail import (
    ConversationInfo as BaseConversationInfo,
    Message as BaseMessage,
    ModAction as BaseModAction,
    UserDossier as BaseUserDossier,
    ConversationAggregate as BaseConversationAggregate,
    UserDossierConversationAggregate as BaseUserDossierConversationAggregate,
    OptionalUserDossierConversationAggregate as BaseOptionalUserDossierConversationAggregate,
)

class ConversationInfo(BaseConversationInfo):
    def __init__(self, d: Mapping[str, Any], client: Client) -> None:
        super().__init__(d)
        self.client: Client = client
        ("")

    async def reply(self, body: str, *, hidden: bool = False, internal: bool = False) -> ConversationAggregate:
        return await self.client.p.modmail.conversation.reply(self.id, body, hidden=hidden, internal=internal)

    async def mark_read(self) -> None:
        await self.client.p.modmail.conversation.mark_read(self.id)

    async def mark_unread(self) -> None:
        await self.client.p.modmail.conversation.mark_unread(self.id)


class Message(BaseMessage):
    def __init__(self, d: Mapping[str, Any], client: Client) -> None:
        super().__init__(d)
        self.client: Client = client
        ("")

class ModAction(BaseModAction):
    pass

class UserDossier(BaseUserDossier):
    pass


@dataclass(repr=False, eq=False, frozen=True)
class ConversationAggregate(BaseConversationAggregate):
    info: ConversationInfo
    messages: Sequence[Message]
    mod_actions: Sequence[ModAction]
    history: Sequence[object]

@dataclass(repr=False, eq=False, frozen=True)
class UserDossierConversationAggregate(ConversationAggregate, BaseUserDossierConversationAggregate):
    user_dossier: UserDossier

@dataclass(repr=False, eq=False, frozen=True)
class OptionalUserDossierConversationAggregate(ConversationAggregate, BaseOptionalUserDossierConversationAggregate):
    user_dossier: Optional[UserDossier]




from __future__ import annotations
from typing import TYPE_CHECKING, Sequence, MutableSequence, Callable, Generic, TypeVar, Iterator, Protocol, Iterable, Optional
if TYPE_CHECKING:
    from ..pagination.paginator import CursorPaginator

import random
import time

from ..pagination.paginator import Resettable
from ..util.ordered_set import BoundedSet
from ..util.event_dispatcher import EventDispatcher


TEvent_contra = TypeVar('TEvent_contra', contravariant=True)
TEvent = TypeVar('TEvent')

class StreamEventHandlerProtocol(Protocol[TEvent_contra]):
    def __call__(self, event: TEvent_contra, /) -> None:
        ...

class StreamEventDispatcher(EventDispatcher[StreamEventHandlerProtocol[TEvent]], Generic[TEvent]):
    def __call__(self, event: TEvent) -> None:
        for handler in self:
            handler(event)


TOutput = TypeVar('TOutput')

class IStandardStreamEventSubject(Iterator[float], Generic[TOutput]):
    output: StreamEventDispatcher[TOutput]
    error: StreamEventDispatcher[Exception]

    def __iter__(self) -> Iterator[float]:
        return self

    def __next__(self) -> float:
        raise NotImplementedError



class Stream(IStandardStreamEventSubject[TOutput]):
    _BASE_POLL_INTERVAL: float = 5.0
    _MAX_POLL_INTERVAL: float = 50.0
    _BACKOFF_FACTOR: float =  2.0
    _JITTER_FACTOR: float = 0.4

    _MEMORY: int = 2000
    _TARGET_LIMIT_MULTIPLIER: float = 1.2

    _BACKTRACK_DEPTH: int = 300

    def __init__(self,
        paginator: CursorPaginator[TOutput],
        extractor: Callable[[TOutput], object],
        *,
        max_limit: int = 100,
        past: Optional[Iterable[TOutput]] = None,
        seen: Optional[Iterable[object]] = None,
    ) -> None:
        if None not in (past, seen):
            raise TypeError("mutually exclusive parameters: `past`, `seen`")
        if not isinstance(paginator, Resettable):
            raise ValueError('paginator must be Resettable')

        self._paginator: CursorPaginator[TOutput] = paginator
        self._paginator__resettable: Resettable = paginator
        self._extractor: Callable[[TOutput], object] = extractor
        self._max_limit: int = max_limit

        init_seen: Iterable[object] = ()
        if seen is not None:
            init_seen = seen
        elif past is not None:
            init_seen = map(extractor, past)
        self._init_seen: Iterable[object] = init_seen

        self._gen: Iterator[float] = self._routine()

        self.output: StreamEventDispatcher[TOutput] = StreamEventDispatcher()
        ("")
        self.error: StreamEventDispatcher[Exception] = StreamEventDispatcher()
        ("")

    def __next__(self) -> float:
        return next(self._gen)

    @staticmethod
    def _intermit(duration: float) -> Iterator[float]:
        point = time.monotonic() + duration
        yield duration
        while (n := point - time.monotonic()) > 0:
            yield n

    def _routine(self) -> Iterator[float]:
        paginator = self._paginator
        paginator__resettable = self._paginator__resettable
        extractor = self._extractor
        max_limit = self._max_limit

        seen: BoundedSet[object] = BoundedSet(self._init_seen, self._MEMORY)
        delay: float = self._BASE_POLL_INTERVAL

        paginator.limit = max_limit

        retrieved: Sequence[TOutput] = ()

        if not paginator.get_cursor():
            while True:
                try:
                    retrieved = paginator.fetch()
                except Exception as error:
                    self.emit_error(error)
                    if delay < self._MAX_POLL_INTERVAL:
                        delay = min(self._BACKOFF_FACTOR * delay, self._MAX_POLL_INTERVAL)
                    yield from self._intermit(random.uniform(
                        delay * (1 - self._JITTER_FACTOR),
                        delay * (1 + self._JITTER_FACTOR)))
                    continue
                break

            delay = self._BASE_POLL_INTERVAL

            paginator__resettable.reset()

            for obj in retrieved:
                seen.add(extractor(obj))

            yield 0

        tuna_limit: float = max_limit
        backtrack_count: int = 0

        while True:
            effective_limit = round(tuna_limit)
            paginator.limit = effective_limit

            retrieved = ()
            fetch_successful: bool = True
            try:
                retrieved = paginator.fetch()
            except Exception as error:
                fetch_successful = False
                self.emit_error(error)

            selection: MutableSequence[TOutput] = []
            for obj in retrieved:
                imprint = self._extractor(obj)
                if imprint not in seen:
                    seen.add(imprint)
                    selection.append(obj)

            for obj in selection:
                self.emit_output(obj)

            backtrack_count += len(selection)

            if (len(selection) < effective_limit) or (backtrack_count > self._BACKTRACK_DEPTH):
                paginator__resettable.reset()
                backtrack_count = 0

            if fetch_successful:
                tuna = tuna_limit
                if not selection:
                    tuna = max(tuna/2, 1)
                elif len(selection) >= effective_limit:
                    tuna = min(2*tuna, max_limit)
                else:
                    tuna = min((self._TARGET_LIMIT_MULTIPLIER * len(selection)), max_limit)
                tuna_limit = tuna

            if fetch_successful:
                delay = self._BASE_POLL_INTERVAL
            else:
                if delay < self._MAX_POLL_INTERVAL:
                    delay = min(self._BACKOFF_FACTOR * delay, self._MAX_POLL_INTERVAL)

            t = 1.
            if not paginator.get_cursor():
                t = random.uniform(
                        delay * (1 - self._JITTER_FACTOR),
                        delay * (1 + self._JITTER_FACTOR))
            yield from self._intermit(t)

    def emit_output(self, output: TOutput) -> None:
        self.output(output)

    def emit_error(self, error: Exception) -> None:
        self.error(error)




from __future__ import annotations
from typing import Mapping, Any

from dataclasses import dataclass

from ..datamemento import DatamementoDataclassesMixin

@dataclass(repr=False, eq=False)
class Widget(DatamementoDataclassesMixin):
    d: Mapping[str, Any]
    idt: str
    kind: str
    header_color: str
    background_color: str
    title: str




from __future__ import annotations
from typing import Any, Mapping, Optional, Sequence

from datetime import datetime, timezone

from ..core.const import AUTHORIZATION_BASE_URL
from ..models.message import CommentMessageCause, MailboxMessage, ComposedMessage, CommentMessage

def load_mailbox_message(d: Mapping[str, Any]) -> MailboxMessage:
    return MailboxMessage(
        d=d,
        subject=d['subject'],
        author_name=d['author'] or '',
        unread=d['new'],
    )

def load_composed_message(d: Mapping[str, Any]) -> ComposedMessage:
    src_user_name: Optional[str] = d['author']
    src_subr_name: Optional[str] = None
    dst_user_name: Optional[str] = None
    dst_subr_name: Optional[str] = None
    dest: str = d['dest']
    if dest.startswith('#'):
        dst_subr_name = dest.lstrip('#')
    else:
        src_subr_name = d['subreddit']
        dst_user_name = dest

    up = load_mailbox_message(d)
    return ComposedMessage(
        d=up.d,
        subject=up.subject,
        author_name=up.author_name,
        unread=up.unread,
        id=int(d['id'], 36),
        unixtime=(unixtime := int(d['created_utc'])),
        datetime=datetime.fromtimestamp(unixtime, timezone.utc),
        body=d['body'],
        body_html=d['body_html'],
        distinguished=d['distinguished'] or '',
        src_user_name=src_user_name,
        src_subr_name=src_subr_name,
        dst_user_name=dst_user_name,
        dst_subr_name=dst_subr_name,
        src_user_id=(
            int(x.partition('_')[2], 36)
            if (x := d['author_fullname']) else
            {
                None: None,
                'reddit': 81524,
                'welcomebot': 404087392163,
            }.get(src_user_name)
        ),
    )

def load_comment_message(d: Mapping[str, Any]) -> CommentMessage:
    up = load_mailbox_message(d)
    return CommentMessage(
        d=up.d,
        subject=up.subject,
        author_name=up.author_name,
        unread=up.unread,
        cause={
            'username_mention': CommentMessageCause.USERNAME_MENTION,
            'post_reply': CommentMessageCause.SUBMISSION_REPLY,
            'comment_reply': CommentMessageCause.COMMENT_REPLY,
        }[d['type']],
        submission=CommentMessage.Submission(
            id=int((context := d['context']).split('/', 5)[4], 36),
            title=d['link_title'],
            comment_count=d['num_comments'],
            permalink_path=(permalink_path := '/'.join(context.split('/', 6)[:-1]) + '/'),
            permalink=AUTHORIZATION_BASE_URL + permalink_path,
        ),
        comment=CommentMessage.Comment(
            id=int(d['id'], 36),
            created_ut=(created_ut := int(d['created_utc'])),
            created_at=datetime.fromtimestamp(created_ut, timezone.utc),
            author_name=d['author'],
            author_id=int(d['author_fullname'].partition('_')[2], 36),
            subreddit_name=d['subreddit'],
            permalink_path=(permalink_path := context.partition('?')[0]),
            permalink=AUTHORIZATION_BASE_URL + permalink_path,
            is_top_level=(is_top_level := (parent_id := d['parent_id']).startswith('t3_')),
            has_parent_comment=(has_parent_comment := not is_top_level),
            parent_comment_id36=(parent_comment_id36 := parent_id.partition('_')[2] if has_parent_comment else ''),
            parent_comment_id=int(parent_comment_id36, 36) if has_parent_comment else 0,
            score=d['score'],
            body=d['body'],
            body_html=d['body_html'],
            voted={False: -1, None: 0, True: 1}[d['likes']],
        ),
    )


def load_composed_message_thread(d: Mapping[str, Any]) -> Sequence[ComposedMessage]:
    first = load_composed_message(d)
    children = []
    if replies := d['replies']:
        children = [load_composed_message(d['data']) for d in replies['data']['children']]
    return [first] + children




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client

from ...models.user_summary import UserSummary
from ...model_loaders.user_summary import load_user_summary
from ...util.base_conversion import to_base36
from ... import http

class GetUserSummary:
    def __init__(self, client: Client) -> None:
        self._client = client

    async def __call__(self, idn: int) -> Optional[UserSummary]:
        id36 = to_base36(idn)
        return await self.by_id36(id36)

    async def by_id36(self, id36: str) -> Optional[UserSummary]:
        full_id36 = 't2_' + id36
        try:
            root = await self._client.request('GET', '/api/user_data_by_account_ids',
                    params={'ids': full_id36})
        except http.exceptions.StatusCodeException as e:
            if e.status_code == 404:
                return None
            raise
        obj_data = root[full_id36]
        return load_user_summary(obj_data, id36)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Iterable

if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.comment_tree_SYNC import SubmissionTreeNode
    from .SYNC import CommentTreeProcedures

from ...model_loaders.comment_tree_SYNC import load_submission_tree_node, load_more_comments_tree_node
from ...exceptions import RejectedResultException
from ... import http
from ...models.comment_tree_SYNC import MoreCommentsTreeNode

class Low:
    def __init__(self, outer: CommentTreeProcedures, client: Client) -> None:
        self._outer = outer
        self._client = client

    def get(self,
        submission_id36: str,
        comment_id36: Optional[str] = None,
        *,
        sort: str = 'confidence',
        limit: Optional[int] = None,
        depth: Optional[int] = None,
        context: Optional[int] = None,
    ) -> Optional[SubmissionTreeNode]:
        try:
            return self.fetch(
                submission_id36,
                comment_id36,
                sort=sort,
                limit=limit,
                depth=depth,
                context=context,
            )
        except http.exceptions.StatusCodeException as e:
            if e.status_code == 404:
                return None
            raise
        except RejectedResultException:
            return None

    def fetch(self,
        submission_id36: str,
        comment_id36: Optional[str] = None,
        *,
        sort: str = 'confidence',
        limit: Optional[int] = None,
        depth: Optional[int] = None,
        context: Optional[int] = None,
    ) -> SubmissionTreeNode:
        tree_node = self.fetch_lenient(
            submission_id36,
            comment_id36,
            sort=sort,
            limit=limit,
            depth=depth,
            context=context,
        )

        if comment_id36 is not None and not tree_node.children:
            raise RejectedResultException('Specified comment not returned.')

        return tree_node

    def fetch_lenient(self,
        submission_id36: str,
        comment_id36: Optional[str] = None,
        *,
        sort: str = 'confidence',
        limit: Optional[int] = None,
        depth: Optional[int] = None,
        context: Optional[int] = None,
    ) -> SubmissionTreeNode:
        def g() -> Iterable[tuple[str, str]]:
            if comment_id36 is not None:
                yield ('comment', comment_id36)
            if sort:
                yield ('sort', sort)
            if limit is not None:
                yield ('limit', str(limit))
            if depth is not None:
                yield ('depth', str(depth))
            if context is not None:
                yield ('context', str(context))

        root = self._client.request('GET', '/comments/' + submission_id36, params=dict(g()))
        return load_submission_tree_node(root, self._client, sort)

    def more_children(self,
        submission_id36: str,
        child_id36s: Iterable[str],
        *,
        sort: str = '',
        depth: Optional[int] = None,
        exact: bool = False,
    ) -> MoreCommentsTreeNode:
        def g() -> Iterable[tuple[str, str]]:
            yield ('link_id', 't3_' + submission_id36)
            yield ('children', ','.join(child_id36s))
            if sort:
                yield ('sort', sort)
            if depth is not None:
                yield ('depth', str(depth))
            if exact:
                yield ('limit_children', '1')

        resp_data = self._client.request('POST', '/api/morechildren', data=dict(g()))
        return load_more_comments_tree_node(resp_data, self._client, submission_id36, sort)




from __future__ import annotations
from typing import Mapping, Any

from ..models.stylesheet import StylesheetInfo, StylesheetImage

def load_stylesheet_info(d: Mapping[str, Any]) -> StylesheetInfo:
    return StylesheetInfo(
        content=d['stylesheet'],
        images=[
            StylesheetImage(
                name=m['name'],
                url=m['url'],
            )
            for m in d['images']
        ],
    )




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.message_SYNC import MailboxMessage, ComposedMessage, CommentMessage

from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.message_sync1 import (
    MessageListingPaginator,
    ComposedMessageListingPaginator,
    CommentMessageListingPaginator,
    ComposedMessageThreadListingPaginator,
)

class Pulls:
    def __init__(self, client: Client) -> None:
        self._client = client

    def inbox(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[MessageListingPaginator, MailboxMessage]:
        p = MessageListingPaginator(self._client, '/message/inbox')
        return ImpartedPaginatorChainingIterator(p, amount)

    def unread(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[MessageListingPaginator, MailboxMessage]:
        p = MessageListingPaginator(self._client, '/message/unread')
        return ImpartedPaginatorChainingIterator(p, amount)

    def messages(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[ComposedMessageThreadListingPaginator, Sequence[ComposedMessage]]:
        p = ComposedMessageThreadListingPaginator(self._client, '/message/messages')
        return ImpartedPaginatorChainingIterator(p, amount)

    def sent(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[ComposedMessageListingPaginator, ComposedMessage]:
        p = ComposedMessageListingPaginator(self._client, '/message/sent')
        return ImpartedPaginatorChainingIterator(p, amount)

    def comment_replies(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[CommentMessageListingPaginator, CommentMessage]:
        p = CommentMessageListingPaginator(self._client, '/message/comments')
        return ImpartedPaginatorChainingIterator(p, amount)

    def post_replies(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[CommentMessageListingPaginator, CommentMessage]:
        p = CommentMessageListingPaginator(self._client, '/message/selfreply')
        return ImpartedPaginatorChainingIterator(p, amount)

    def mentions(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[CommentMessageListingPaginator, CommentMessage]:
        p = CommentMessageListingPaginator(self._client, '/message/mentions')
        return ImpartedPaginatorChainingIterator(p, amount)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Any, Iterable
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.custom_feed_SYNC import CustomFeed

import json

from ...model_loaders.custom_feed_SYNC import load_custom_feed
from ... import exceptions
from ... import http
from ...iterators.chunking import chunked
from ...iterators.call_chunk_calling_iterator import CallChunkCallingIterator
from ...iterators.call_chunk import CallChunk

class CustomFeedProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client
        self._json_encoder = encoder = json.JSONEncoder()
        self._json_encode = encoder.encode

    def fetch(self, user: str, feed: str) -> CustomFeed:
        """Fetch a custom feed resource.

        .. .PARAMETERS

        :param `str` user:
            A user name.
        :param `str` feed:
            Custom feed name.

        .. .RETURNS

        :rtype: :class:`~.models.custom_feed_SYNC.CustomFeed`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `MULTI_NOT_FOUND`:
                The custom feed specified doesn't exist or you don't have permission to view it.
        """
        root = self._client.request('GET', f'/api/multi/user/{user}/m/{feed}')
        return load_custom_feed(root['data'], self._client)

    def get(self, user: str, feed: str) -> Optional[CustomFeed]:
        """Get a custom feed resource.

        Returns `None` if the custom feed was not found.

        .. .PARAMETERS

        :param `str` user:
            A user name.
        :param `str` feed:
            Custom feed name.

        .. .RETURNS

        :rtype: `Optional`\\[:class:`~.models.custom_feed_SYNC.CustomFeed`]
        """
        try:
            root = self._client.request('GET', f'/api/multi/user/{user}/m/{feed}')
        except exceptions.RedditError as e:
            if e.label == 'MULTI_NOT_FOUND':
                return None
            raise
        return load_custom_feed(root['data'], self._client)

    def retrieve_mine(self) -> Sequence[CustomFeed]:
        """Fetch a list of custom feeds curated by the current user.

        .. .RETURNS

        :rtype: `Sequence`\\[:class:`~.models.custom_feed_SYNC.CustomFeed`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        result = self._client.request('GET', '/api/multi/mine')
        return [load_custom_feed(d['data'], self._client) for d in result]

    def retrieve_yours(self, user: str) -> Sequence[CustomFeed]:
        """Fetch a list of custom feeds curated by a given user.

        .. .PARAMETERS

        :param `Optional[str]` user:
            The user in which to retrieve custom feeds for.

        .. .RETURNS

        :rtype: `Sequence`\\[:class:`~.models.custom_feed_SYNC.CustomFeed`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_DOESNT_EXIST`:
                The specified user does not exist.
        """
        result = self._client.request('GET', f'/api/multi/user/{user}')
        return [load_custom_feed(d['data'], self._client) for d in result]

    def retrieve(self, user: Optional[str] = None) -> Sequence[CustomFeed]:
        """Fetch a list of custom feeds.

        Calls :meth:`.retrieve_yours` if the parameter is specified,
        else calls :meth:`.retrieve_mine`.
        """
        if user is None:
            return self.retrieve_mine()
        return self.retrieve_yours(user)

    def create(self,
        user: str,
        feed: str,
        *,
        title: Optional[str] = None,
        description: Optional[str] = None,
        subreddit_names: Sequence[str] = (),
        exposed: bool = False,
    ) -> CustomFeed:
        """Create a custom feed.

        .. .PARAMETERS

        :param `str` user:
        :param `str` feed:
        :param `Optional[str]` title:
            Title.

            No longer than 50 characters.

            If null, feed name is used.
        :param `Optional[str]` description:
            Markdown description text.

            Defaults to empty string.
        :param `Sequence[str]` subreddit_names:
            A list of subreddit names to include in the custom feed.
        :param `bool` exposed:
            Whether the custom feed is publicly accessible.

        .. .RETURNS

        :returns: The newly created custom feed.
        :rtype: :class:`~.models.custom_feed_SYNC.CustomFeed`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `MULTI_CANNOT_EDIT`:
               - The username specified does not exist.
               - You don't have permission to create a custom feed there.

            + `MULTI_EXISTS`:
                The custom feed already exists.
            + `TOO_LONG`:
                The text specified by `title` is over 50 characters.
        """
        json_data: dict[str, Any] = {}
        if title is not None:
            json_data['display_name'] = title
        if description is not None:
            json_data['description_md'] = description
        if subreddit_names:
            json_data['subreddits'] = {'name': nm for nm in subreddit_names}
        if not exposed:
            json_data['visibility'] = 'public'

        json_str = self._json_encode(json_data)
        root = self._client.request('POST', f'/api/multi/user/{user}/m/{feed}', data={'model': json_str})
        return load_custom_feed(root['data'], self._client)

    def put(self,
        user: str,
        feed: str,
        *,
        title: Optional[str] = None,
        description: Optional[str] = None,
        subreddit_names: Sequence[str] = (),
        exposed: bool = False,
    ) -> CustomFeed:
        """Create or update a custom feed.

        Behaves similar to :meth:`.create`.

        .. .PARAMETERS

        :param `str` user:
        :param `str` feed:
        :param `Optional[str]` title:
        :param `Optional[str]` description:
        :param `Sequence[str]` subreddit_names:
        :param `bool` exposed:

        .. .RETURNS

        :returns: The created/updated created custom feed.
        :rtype: :class:`~.models.custom_feed_SYNC.CustomFeed`

        .. .RAISES

        :(raises): Similar to :meth:`.create`,
            but `RedditError('MULTI_EXISTS')` is not possible.
        """
        json_data: dict[str, Any] = {}
        if title is not None:
            json_data['display_name'] = title
        if description is not None:
            json_data['description_md'] = description
        if subreddit_names:
            json_data['subreddits'] = [{'name': nm} for nm in subreddit_names]
        if not exposed:
            json_data['visibility'] = 'public'

        json_str = self._json_encode(json_data)
        root = self._client.request('PUT', f'/api/multi/user/{user}/m/{feed}', data={'model': json_str})
        return load_custom_feed(root['data'], self._client)

    def delete(self, user: str, feed: str) -> None:
        """Delete a custom feed.

        .. .PARAMETERS

        :param `str` user:
        :param `str` feed:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `MULTI_NOT_FOUND`:
               - The custom feed name specified does not exist.
               - The username specified does not exist.

            + `MULTI_CANNOT_EDIT`:
                You don't have permission to delete the specified custom feed
                because it does not belong to you.
        """
        self._client.request('DELETE', f'/api/multi/user/{user}/m/{feed}')

    def copy(self, src_user: str, src_feed: str, dst_user: str, dst_feed: str, *,
            title: Optional[str] = None, description: Optional[str] = None) -> CustomFeed:
        """Copy a custom feed.

        The description for the new custom feed will be like "`copied from u/spez`" etc.
        unless overridden by the `description` parameter.

        .. .PARAMETERS

        :param `str` src_user:
        :param `str` src_feed:
        :param `str` dst_user:
        :param `str` dst_feed:

        .. .RETURNS

        :returns: The newly created custom feed.
        :rtype: :class:`~.models.custom_feed_SYNC.CustomFeed`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `MULTI_NOT_FOUND`:
                The source custom feed was not found.
            + `BAD_MULTI_PATH`:
                The destination path was not valid.
            + `MULTI_EXISTS`:
                The destination custom feed already exists.
            + `TOO_LONG`:
                The text specified by `title` is over 50 characters.
        """
        data = {
            'from': f'/user/{src_user}/m/{src_feed}',
            'to': f'/user/{dst_user}/m/{dst_feed}',
        }
        if title is not None:
            data['display_name'] = title
        if description is not None:
            data['description_md'] = description

        root = self._client.request('POST', '/api/multi/copy', data=data)
        return load_custom_feed(root['data'], self._client)

    def contains(self, user: str, feed: str, sr_name: str) -> bool:
        """Tell if a subreddit is in a custom feed.

        .. .PARAMETERS

        :param `str` user:
        :param `str` feed:
        :param `str` sr_name:
            Subreddit name.

        .. .RETURNS

        :rtype: `bool`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `MULTI_NOT_FOUND`:
                The custom feed does not exist.
        """
        try:
            self._client.request('GET', f'/api/multi/user/{user}/m/{feed}/r/{sr_name}')
        except http.exceptions.StatusCodeException as e:
            if e.status_code == 500:
                return False
            raise
        except exceptions.RedditError as e:
            if e.label == 'SUBREDDIT_NOEXIST':
                return False
            raise
        return True

    def add_item(self, user: str, feed: str, sr_name: str) -> None:
        """Add a subreddit to a custom feed.

        The `user` parameter does not have to match the current user's name!
        If the user name refers to a user that exists and the feed name you
        specify exists on that user, you'll get a `MULTI_CANNOT_EDIT` API error.
        If the username or feed name you specify doesn't exist, the custom feed
        will be created, seemingly under that user's name, but the custom feed
        will only be visible to you.

        If the specified feed name doesn't exist, it will be created.

        .. .PARAMETERS

        :param `str` user:
        :param `str` feed:
        :param `str` sr_name:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `SUBREDDIT_NOEXIST`:
                The specified subreddit does not exist.
            + `MULTI_CANNOT_EDIT`:
                The target custom feed exists and you don't have permission to add to it.
        """
        json_str = self._json_encode({"name": "abc"})
        self._client.request('PUT', f'/api/multi/user/{user}/m/{feed}/r/{sr_name}', data={'model': json_str})

    def bulk_add_items(self, user: str, feed: str, sr_names: Iterable[str]) -> CallChunkCallingIterator[None]:
        """Bulk add subreddits to a custom feed.

        If any of the subreddit names in `sr_names` doesn't exist, the request will fail
        with a 500 HTTP error and none of the subreddits will be added.

        .. .PARAMETERS

        :param `str` user:
        :param `str` feed:
        :param `Iterable[str]` sr_names:

        .. .RETURNS

        :rtype: `~.iterators.call_chunk_calling_iterator.CallChunkCallingIterator`\\[`None`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `500`:
               - The specified username does not exist.
               - The specified custom feed name doesn't exist.
               - One of the subreddits specified in the `sr_names` list does not exist.
        """
        def mass_add_items(sr_names: Sequence[str]) -> None:
            data = {'path': f'/user/{user}/m/{feed}', 'sr_names': (','.join(sr_names))}
            self._client.request('POST', '/api/multi/add_srs_bulk', data=data)

        return CallChunkCallingIterator(CallChunk(mass_add_items, chunk) for chunk in chunked(sr_names, 300))

    def remove_item(self, user: str, feed: str, sr_name: str) -> None:
        """Remove a subreddit from a custom feed.

        .. .PARAMETERS

        :param `str` user:
        :param `str` feed:
        :param `str` sr_name:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `MULTI_NOT_FOUND`:
               - The specified username does not exist.
               - The specified custom feed name does not exist.
            + `MULTI_CANNOT_EDIT`:
                You don't have permission.
        """
        self._client.request('DELETE', f'/api/multi/user/{user}/m/{feed}/r/{sr_name}')




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, cast, Mapping
if TYPE_CHECKING:
    from collections.abc import AsyncIterator

import asyncio

# https://pypi.org/project/websockets/
import websockets  # type: ignore[import]
import websockets.legacy.client  # type: ignore[import]
import websockets.typing  # type: ignore[import]
import websockets.exceptions  # type: ignore[import]

from ..reg_ASYNC import register
from ...websocket_ASYNC import PartiallyImplementedWebSocket, DEFAULT_WAITTIME
from ... import exceptions
from ... import events
from ...events import Frame, TextMessage, BytesMessage
from ...const import Side, ConnectionState

class WebSocketClient(PartiallyImplementedWebSocket):
    SIDE: int = Side.CLIENT

    def __init__(self, ws: websockets.legacy.client.WebSocketClientProtocol) -> None:
        super().__init__()
        self.ws: websockets.legacy.client.WebSocketClientProtocol = ws
        ("")

    async def _send_frame_impl(self, m: Frame) -> None:
        try:
            await self.ws.write_frame(opcode=m.opcode, data=m.data, fin=m.fin)
        except Exception as cause:
            raise exceptions.TransportError from cause

    async def _pulse_impl(self, *, timeout: float = -2) -> AsyncIterator[object]:
        t: Optional[float] = timeout
        if timeout == -2:
            t = self.default_waittime
        elif timeout == -1:
            t = None
        elif timeout < 0:
            raise ValueError(f'invalid timeout value: {t}')

        try:
            data = await asyncio.wait_for(self.ws.recv(), t)
        except asyncio.TimeoutError as cause:
            raise exceptions.TimeoutException from cause
        except websockets.exceptions.ConnectionClosed as e:
            self.close_code: int = e.code
            self.close_reason: str = e.reason
            self.state = ConnectionState.CLOSED
            yield events.ConnectionClosed()
        except Exception as cause:
            raise exceptions.TransportError from cause
        else:
            if isinstance(data, str):
                yield TextMessage(data)
            else:
                yield BytesMessage(data)

    async def close(self, code: Optional[int] = 1000, reason: str = '', *, waitfor: float = -2) -> None:
        if code is None:
            raise RuntimeError('must specify a close code with websockets library websockets')

        t: Optional[float] = waitfor
        if waitfor == -2:
            t = self.default_waittime
        elif waitfor == -1:
            t = None
        elif waitfor < 0:
            raise ValueError(f'invalid waitfor value: {waitfor}')

        if t is None:
            raise RuntimeError('the websockets library does not support infinite waitfor')

        if self.state == ConnectionState.CLOSED:
            return
        if self.state != ConnectionState.OPEN:
            raise exceptions.InvalidStateException(f'operation not permitted in {self.state.name!r} state')

        self.state: ConnectionState = ConnectionState.CLOSING

        self.ws.close_timeout = t
        try:
            await self.ws.close(code, reason)
        except Exception as cause:
            raise exceptions.TransportError from cause

        if self.close_code < 0:
            self.close_code = 1006
        self.state = ConnectionState.CLOSED


async def connect(
    url: str,
    *,
    subprotocols: Sequence[str] = (),
    headers: Optional[Mapping[str, str]] = None,
    timeout: float = -2,
) -> WebSocketClient:
    t: Optional[float] = timeout
    if timeout == -2:
        t = DEFAULT_WAITTIME
    elif timeout == -1:
        t = None
    elif timeout < 0:
        raise ValueError(f'invalid timeout value: {timeout}')

    subp = cast(Optional[Sequence[websockets.typing.Subprotocol]], subprotocols if subprotocols else None)
    coro = websockets.legacy.client.connect(
        url,
        extra_headers=headers,
        subprotocols=subp,
    )
    try:
        ws = await asyncio.wait_for(coro, t)
    except asyncio.TimeoutError as cause:
        raise exceptions.TimeoutException from cause
    except Exception as cause:
        raise exceptions.TransportError from cause

    wsc = WebSocketClient(ws)
    wsc.subprotocol = ws.subprotocol or ''
    return wsc


name: str = websockets.__name__
version: str = websockets.__version__  # type: ignore[attr-defined]
register(
    adaptor_module_name=__name__,
    name=name,
    version=version,
    connect=connect,
)




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Sequence
if TYPE_CHECKING:
    from ....client_SYNC import Client

from ..listing.mixins.sort_SYNC import Sort
from ..listing.listing_paginator import ListingPaginator
from ....models.user_SYNC import User
from ....model_loaders.user_SYNC import load_user

class UserSearchPaginator(
    Sort[User],
    ListingPaginator[User],
):
    def __init__(self,
        client: Client,
        url: str,
        query: str,
    ) -> None:
        super().__init__(client, url)
        self.query: str = query
        ("")

    def _generate_params(self) -> Iterable[tuple[str, str]]:
        yield from super()._generate_params()
        yield ('q', self.query)

    def fetch(self) -> Sequence[User]:
        data = self._fetch_data()
        return [load_user(d['data'], self.client) for d in data['children']]




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, MutableMapping
if TYPE_CHECKING:
    from ..http.handler_ASYNC import Handler

from .const import RESOURCE_BASE_URL, TOKEN_OBTAINMENT_URL, TRUSTED_ORIGINS
from ..http.http_client_ASYNC import HTTPClient as BaseHTTPClient
from ..auth.types import AuthorizationGrant
from .authorizer_ASYNC import Authorized
from .rate_limited_ASYNC import RateLimited
from .recorded_ASYNC import Recorded
from .reddit_please_send_json_ASYNC import RedditPleaseSendJSON
from .reddit_token_obtainment_client_ASYNC import RedditTokenObtainmentClient
from ..http.transport.reg_ASYNC import new_connector
from ..http.util.case_insensitive_dict import CaseInsensitiveDict
from ..http.misc.apply_params_and_headers_ASYNC import ApplyDefaultHeaders
from .user_agent_ASYNC import get_user_agent
from ..auth.token import Token
from .recorded_ASYNC import Last
from .authorizer_ASYNC import Authorizer
from ..http.send_params import SendParams
from ..http.exchange import Exchange
from .direct_by_origin_ASYNC import DirectByOrigin


DEFAULT_TIMEOUT: float = 8


class HTTPClient(BaseHTTPClient):
    """An HTTP client specialised for the purposes of this library."""

    @property
    def last(self) -> Last:
        return self._last

    def __init__(self, handler: Handler, *,
        headers: Optional[MutableMapping[str, str]] = None,
    ) -> None:
        recorder = Recorded(handler)
        super().__init__(recorder)
        self.base_url: str = RESOURCE_BASE_URL
        ("")
        self.timeout: float = DEFAULT_TIMEOUT
        ("")
        self.user_agent_base: str = ''
        ("")
        self.headers: MutableMapping[str, str] = CaseInsensitiveDict() if headers is None else headers
        ("")
        self._last: Last = Last(recorder)

    async def _send(self, p: SendParams) -> Exchange:
        (_d := p.requisition.headers).update({**self.headers, **_d})
        return await super()._send(p)

    def get_user_agent(self) -> str:
        return self.headers.get('User-Agent', '')

    def set_user_agent(self, value: Optional[str]) -> None:
        if value is None:
            self.headers.pop('User-Agent', None)
        else:
            self.headers['User-Agent'] = value


class RedditHTTPClient(HTTPClient):
    """An HTTP client for making requests to the public Reddit API."""

    @property
    def authorizer(self) -> Authorizer:
        return self.fetch_authorizer()

    def __init__(self, handler: Handler, *,
        headers: Optional[MutableMapping[str, str]] = None,
        authorizer: Optional[Authorizer] = None,
    ) -> None:
        super().__init__(handler, headers=headers)
        self._authorizer: Optional[Authorizer] = authorizer

    def fetch_authorizer(self) -> Authorizer:
        if self._authorizer is None:
            raise RuntimeError('value not set')
        return self._authorizer

    def fast_set_authorizer(self, value: Optional[Authorizer]) -> None:
        """Changes the value of `self.authorizer`.

        Note, the :attr:`authorizer` attribute is just a holder and changing
        its value will not change the underlying authorizer.
        """
        self._authorizer = value


def build_reddit_http_client(
    client_id: str,
    client_secret: str,
    grant: AuthorizationGrant,
) -> RedditHTTPClient:
    connector = new_connector()
    ua = get_user_agent(module_member=connector)
    headers = CaseInsensitiveDict({'User-Agent': ua})
    token_client = RedditTokenObtainmentClient(
        HTTPClient(ApplyDefaultHeaders(connector, headers)),
        TOKEN_OBTAINMENT_URL,
        (client_id, client_secret),
        grant,
    )
    authorizer = Authorizer(token_client)
    handler: Handler
    handler = RedditPleaseSendJSON(RateLimited(Authorized(connector, authorizer)))
    directions = {x: handler for x in TRUSTED_ORIGINS}
    handler = DirectByOrigin(connector, directions)
    http = RedditHTTPClient(handler, headers=headers, authorizer=authorizer)
    http.user_agent_base = ua
    return http

def build_reddit_http_client_from_access_token(
    access_token: str,
) -> RedditHTTPClient:
    connector = new_connector()
    ua = get_user_agent(module_member=connector)
    headers = CaseInsensitiveDict({'User-Agent': ua})
    authorizer = Authorizer(token=Token(access_token))
    handler: Handler
    handler = RedditPleaseSendJSON(RateLimited(Authorized(connector, authorizer)))
    directions = {x: handler for x in TRUSTED_ORIGINS}
    handler = DirectByOrigin(connector, directions)
    http = RedditHTTPClient(handler, headers=headers, authorizer=authorizer)
    http.user_agent_base = ua
    return http




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any, Sequence
if TYPE_CHECKING:
    from ..client_SYNC import Client

from .submission_collection import (
    SubmissionCollectionInfo as BaseSubmissionCollectionInfo,
    SubmissionCollection as BaseSubmissionCollection,
)
from .submission_SYNC import Submission
from ..model_loaders.submission_SYNC import load_submission

class SubmissionCollectionInfo(BaseSubmissionCollectionInfo):
    def __init__(self, d: Mapping[str, Any], client: Client) -> None:
        super().__init__(d)
        self.client: Client = client
        ("")

    def add_post(self, submission_id: int) -> None:
        self.client.p.collection.add_post(self.uuid, submission_id)

    def remove_post(self, submission_id: int) -> None:
        self.client.p.collection.remove_post(self.uuid, submission_id)

    def reorder(self, submission_ids: Sequence[int]) -> None:
        self.client.p.collection.reorder(self.uuid, submission_ids)

    def delete(self) -> None:
        self.client.p.collection.delete(self.uuid)

    def set_title(self, title: str) -> None:
        self.client.p.collection.set_title(self.uuid, title)

    def set_description(self, desc: str) -> None:
        self.client.p.collection.set_description(self.uuid, desc)

class SubmissionCollection(SubmissionCollectionInfo, BaseSubmissionCollection):
    @property
    def submissions(self) -> Sequence[Submission]:
        return self.__submissions

    def __init__(self, d: Mapping[str, Any], client: Client) -> None:
        super().__init__(d, client)
        load = lambda d: load_submission(d, client)
        self.__submissions: Sequence[Submission] = self._load_submissions(d, load)




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping
if TYPE_CHECKING:
    from ..client_SYNC import Client

from ..models.user_SYNC import User, SuspendedUser


def load_user(d: Mapping[str, Any], client: Client) -> User:
    return User(d, client)

def load_potentially_suspended_user(d: Mapping[str, Any], client: Client) -> object:
    if d.get('is_suspended', False):
        return SuspendedUser(d, client)
    return User(d, client)




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, ClassVar
if TYPE_CHECKING:
    from ..http.handler_ASYNC import Handler
    from ..http.send_params import SendParams
    from ..http.exchange import Exchange

from ..http.delegating_handler_ASYNC import DelegatingHandler

class RedditPleaseSendJSON(DelegatingHandler):
    PARAMS: ClassVar[Mapping[str, str]] = {
        'raw_json': '1',
        'api_type': 'json',
    }

    def __init__(self, handler: Handler) -> None:
        super().__init__(handler)
        self.params: Mapping[str, str] = dict(self.PARAMS)
        ("")

    async def _send(self, p: SendParams) -> Exchange:
        params = p.requisition.params
        for k, v in self.params.items():
            if params.setdefault(k, v) == '\0':
                del params[k]
        return await super()._send(p)




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any
if TYPE_CHECKING:
    from ...http.http_client_ASYNC import HTTPClient
    from ...types import JSON_ro

from ...http.util.json_loading import load_json_from_response_but_prefer_status_code_exception_on_failure
from .token import Token
from ...auth.utils import apply_basic_auth
from .exceptions import raise_for_token_server_response_error
from ..core.const import TOKEN_OBTAINMENT_URL, REDDIT_MOBILE_IOS_CLIENT_ID, GRANT_DATA

class TokenObtainmentClient:
    def __init__(self, http: HTTPClient, url: str,
            client_id: str,
            grant_data: JSON_ro) -> None:
        self.http: HTTPClient = http
        ("")
        self.url: str = url
        ("")
        self.client_id: str = client_id
        ("")
        self.grant_data: JSON_ro = grant_data
        ("")

    async def fetch_data(self) -> Mapping[str, Any]:
        headers: dict[str, str] = {}
        apply_basic_auth(headers, self.client_id, '')
        resp = await self.http.request('POST', self.url, headers=headers, json=self.grant_data)

        json_data = load_json_from_response_but_prefer_status_code_exception_on_failure(resp)
        raise_for_token_server_response_error(json_data)
        resp.ensure_successful_status()
        return json_data

    async def fetch_token(self) -> Token:
        return Token.from_dict(await self.fetch_data())


def new_token_obtainment_client(http: HTTPClient) -> TokenObtainmentClient:
    return TokenObtainmentClient(
        http,
        url=TOKEN_OBTAINMENT_URL,
        client_id=REDDIT_MOBILE_IOS_CLIENT_ID,
        grant_data=GRANT_DATA,
    )




from __future__ import annotations
from typing import Sequence

from .listing_async_paginator import ListingAsyncPaginator
from ....models.comment_ASYNC import LooseComment
from ....model_loaders.comment_ASYNC import load_loose_comment

class LooseCommentListingAsyncPaginator(ListingAsyncPaginator[LooseComment]):
    async def fetch(self) -> Sequence[LooseComment]:
        data = await self._fetch_data()
        return [load_loose_comment(d['data'], self.client) for d in data['children']]



"""OAuth2 utilities.

This module aims to provide general-purpose OAuth2 utilities.
"""

from . import token  # noqa: F401
from .token import Token as Token  # noqa: F401
from . import exceptions  # noqa: F401
from . import grants  # noqa: F401
from . import types  # noqa: F401
from . import utils  # noqa: F401




from __future__ import annotations
from typing import Any, Mapping

from ..models.user_relationship import UserRelationship, FriendRelationship, BannedSubredditUserRelationship

def load_user_relationship(d: Mapping[str, Any]) -> UserRelationship:
    return UserRelationship(d)

def load_friend_relationship(d: Mapping[str, Any]) -> FriendRelationship:
    return FriendRelationship(d)

def load_banned_subreddit_user_relation(d: Mapping[str, Any]) -> BannedSubredditUserRelationship:
    return BannedSubredditUserRelationship(d)




from __future__ import annotations
from typing import Iterable

from ..listing.mixins.sort_ASYNC import Sort
from ..listing.mixins.subreddit_detail_ASYNC import SubredditDetail
from ..listing.submission_listing_async_paginator import SubmissionListingAsyncPaginator
from ..listing.comment_listing_async_paginator import LooseCommentListingAsyncPaginator
from ..listing.submission_and_comment_listing_async_paginator import SubmissionAndLooseCommentListingAsyncPaginator
from ....models.comment_ASYNC import LooseComment
from ....models.submission_ASYNC import Submission

class OverviewListingAsyncPaginator(
    Sort[object],
    SubredditDetail[object],
    SubmissionAndLooseCommentListingAsyncPaginator,
): pass

class CommentsListingAsyncPaginator(
    Sort[LooseComment],
    SubredditDetail[LooseComment],
    LooseCommentListingAsyncPaginator,
): pass

class SubmittedListingAsyncPaginator(
    Sort[Submission],
    SubredditDetail[Submission],
    SubmissionListingAsyncPaginator,
): pass

class GildedListingAsyncPaginator(
    Sort[object],
    SubredditDetail[object],
    SubmissionAndLooseCommentListingAsyncPaginator,
): pass

class UpvotedListingAsyncPaginator(
    Sort[Submission],
    SubredditDetail[Submission],
    SubmissionListingAsyncPaginator,
): pass

class DownvotedListingAsyncPaginator(
    Sort[Submission],
    SubredditDetail[Submission],
    SubmissionListingAsyncPaginator,
): pass

class HiddenListingAsyncPaginator(
    Sort[Submission],
    SubredditDetail[Submission],
    SubmissionListingAsyncPaginator,
): pass

class SavedListingAsyncPaginator(
    Sort[object],
    SubredditDetail[object],
    SubmissionAndLooseCommentListingAsyncPaginator,
): pass

class SavedSubmissionsListingAsyncPaginator(
    Sort[Submission],
    SubredditDetail[Submission],
    SubmissionListingAsyncPaginator,
):
    def _generate_params(self) -> Iterable[tuple[str, str]]:
        yield from super()._generate_params()
        yield ('type', 'links')

class SavedCommentsListingAsyncPaginator(
    Sort[LooseComment],
    SubredditDetail[LooseComment],
    LooseCommentListingAsyncPaginator,
):
    def _generate_params(self) -> Iterable[tuple[str, str]]:
        yield from super()._generate_params()
        yield ('type', 'comments')
